classdef DOFDMDemodulator < comm.internal.OFDMBase
%OFDMDemodulator Apply OFDM demodulation to the input signal.
%   H = comm.OFDMDemodulator returns an orthogonal frequency-division
%   multiplexing (OFDM) demodulator System object, H. This object applies
%   OFDM demodulation to a serialized time-domain input signal to obtain a
%   frequency-domain output signal.
% 
%   H = comm.OFDMDemodulator(Name,Value) creates an OFDM demodulator System
%   object, H, with the specified property Name set to the specified Value.
%   You can specify additional name-value pair arguments in any order as
%   (Name1,Value1,...,NameN,ValueN).
% 
%   H = comm.OFDMDemodulator(G) creates an OFDM demodulator System object,
%   H, from an OFDM modulator System object G. Specifically,
%       H.FFTLength            = G.FFTLength
%       H.NumGuardBandCarriers = G.NumGuardBandCarriers
%       H.RemoveDCCarrier      = G.InsertDCNull
%       H.PilotOutputPort      = G.PilotInputPort
%       H.PilotCarrierIndices  = G.PilotCarrierIndices
%       H.CyclicPrefixLength   = G.CyclicPrefixLength
%       H.NumSymbols           = G.NumSymbols
%       H.NumReceiveAntennas   = 1
% 
%   Step method syntax:
%
%   Y = step(H, X) applies OFDM demodulation to the input signal X and
%   returns the result in Y. The input X is a double precision, real or
%   complex, 2D matrix of size ((NFFT+CPLen)*Nsym) x Nr if CPLen is a
%   scalar or of size (NFFT*Nsym+sum(CPLen)) x Nr if CPLen is a row vector.
%   NFFT is the number of frequency subcarriers (the FFTLength property
%   value of H), Nsym is the number of OFDM symbols per antenna (the
%   NumSymbols property value of H), CPLen is the cyclic prefix length(s)
%   (the CyclicPrefixLength property value of H), and Nr is the number of
%   receive antennas (the NumReceiveAntennas property value of H). The
%   output signal Y is a double precision, complex, 3D array of size Nf x
%   Nsym x Nr, where Nf is the number of frequency subcarriers excluding
%   guard bands and DC null, if applicable. Mathematically, Nf =
%   H.FFTLength - sum(H.NumGuardBandCarriers) - H.RemoveDCCarrier. Note
%   that Y becomes a 2D matrix when Nr is 1.
%   
%   [DATA, PILOT] = step(H, X) separates the PILOT signal on the frequency
%   subcarriers specified by the PilotCarrierIndices property value of H,
%   denoted by pIndex, after OFDM demodulation. This syntax applies when
%   you set the PilotOutputPort property of H to true. The output PILOT is
%   a double precision, complex, 3D array of size Np x Nsym x Nr, where Np
%   is the number of pilot subcarriers in each symbol, i.e., size(pIndex,
%   1). The output DATA is a double precision, complex, 3D array of size Nd
%   x Nsym x Nr, where Nd is the number of data subcarriers in each symbol.
%   Mathematically, Nd = Nf - numel(pIndex)/size(pIndex, 2). Refer to the
%   help of the PilotCarrierIndices property for more information about how
%   Nd is calculated.
% 
%   OFDMDemodulator methods:
%
%   step     - Apply OFDM demodulation to the input signal (see above)
%   release  - Allow property value and input characteristics changes
%   clone    - Create OFDM demodulation object with same property values
%   isLocked - Locked status (logical)
%   info     - Return input and output size information of the object
%   showResourceMapping - Display subcarrier mapping for data, pilot,
%                         guard band and null signals
%
%   OFDMDemodulator properties:
%
%   FFTLength            - FFT length, i.e., the number of subcarriers
%   NumGuardBandCarriers - Number of left and right guard band subcarriers
%   RemoveDCCarrier      - Remove the DC subcarrier from the output
%   PilotOutputPort      - Enable a secondary output for pilot signals
%   PilotCarrierIndices  - Subcarrier indices for pilot signals on each OFDM symbol
%   CyclicPrefixLength   - Cyclic prefix length(s)
%   NumSymbols           - Number of symbols at each receive antenna
%   NumReceiveAntennas   - Number of receive antennas
% 
%   % Example 1: 
%   %   OFDM modulation and demodulation through a single-input 
%   %   single-output (SISO) channel with data only input/output.
%     
%   % Create an OFDM demodulator System object for 10 symbols
%   hDemod = comm.OFDMDemodulator('NumSymbols', 10);
%   % Create an OFDM modulator System object from hDemod
%   hMod = comm.OFDMModulator(hDemod);
%     
%   % Obtain the input and output dimension information via the info method
%   modDim = info(hMod);
%     
%   % Create a random input to the modulator object based on the input
%   % dimension information
%   modIn = complex(rand(modDim.DataInputSize), rand(modDim.DataInputSize));
%    
%   % Apply OFDM modulation to the random input
%   modOut = step(hMod, modIn);
%   % Assume the channel is noiseless and apply OFDM demodulation
%   demodOut = step(hDemod, modOut);
%     
%   % Verify that the input to the modulator and output of the demodulator
%   % are the same, within a numeric tolerance of 1e-10
%   isSame = (max(abs(modIn(:) - demodOut(:))) < 1e-10)
%     
%   % Example 2:
%   %   OFDM modulation and demodulation through a single-input 
%   %   single-output (SISO) channel with data and pilot inputs/outputs.
%   
%   % Create an OFDM demodulator System object that has different pilot
%   % subcarrier indices and cyclic prefix lengths for its 3 symbols
%   hDemod = comm.OFDMDemodulator( ...
%       'NumGuardBandCarriers', [9;8], ...
%       'RemoveDCCarrier',      true, ...
%       'PilotOutputPort',      true, ...
%       'PilotCarrierIndices',  [[12;26;40;54],[14;28;38;52],[12;26;40;54]], ...
%       'CyclicPrefixLength',   [16 32 16], ...
%       'NumSymbols',           3);
%   % Create an OFDM modulator System object from hDemod
%   hMod = comm.OFDMModulator(hDemod);
%   
%   % Check subcarrier mapping for data, pilot, guard band and null signals
%   showResourceMapping(hDemod);
%   % Obtain the input and output dimension information via the info method
%   modDim = info(hMod);
%  
%   % Create random data and pilot inputs to the OFDM modulator object
%   dataIn  = complex(rand(modDim.DataInputSize),  rand(modDim.DataInputSize));
%   pilotIn = complex(rand(modDim.PilotInputSize), rand(modDim.PilotInputSize));
%     
%   % Apply OFDM modulation to the random inputs
%   modOut = step(hMod, dataIn, pilotIn);
%   % Assume the channel is noiseless and apply OFDM demodulation
%   [dataOut, pilotOut] = step(hDemod, modOut);
%   
%   % Verify the data and pilots are unchanged through this process
%   isSame = (max(abs([dataIn(:) - dataOut(:); pilotIn(:) - pilotOut(:)])) < 1e-10)
%   
%   % Example 3:
%   %   OFDM modulation and demodulation through a multiple-input 
%   %   multiple-output (MIMO) channel with data and pilot inputs/outputs.
%  
%   % Create an OFDM demodulator System object for 2 transmit and 4 
%   % receive antennas. It has different pilot subcarrier indices for
%   % different transmit antennas
%   hDemod = comm.OFDMDemodulator( ...
%       'FFTLength',            128, ...
%       'NumGuardBandCarriers', [24;23], ...
%       'RemoveDCCarrier',      true, ...
%       'PilotOutputPort',      true, ...
%       'PilotCarrierIndices',  cat(3,[41;53;77;89],[39;55;75;87]), ...
%       'CyclicPrefixLength',   32, ...
%       'NumSymbols',           12, ...
%       'NumReceiveAntennas',   4);
%   % Construct an OFDM modulator System object from hDemod. The
%   % NumTransmitAntennas property of the modulator object is derived from 
%   % the PilotCarrierIndices property of hDemod.
%   hMod = comm.OFDMModulator(hDemod);
%   
%   % Check subcarrier mapping for each transmit antenna with DC at index 0
%   hDemod.showResourceMapping([-64 63]);
%   % Obtain the input and output dimension information via the info method
%   modDim = info(hMod);
% 
%   % Create random data and pilot inputs to the OFDM modulator object
%   dataIn  = complex(rand(modDim.DataInputSize),  rand(modDim.DataInputSize));
%   pilotIn = complex(rand(modDim.PilotInputSize), rand(modDim.PilotInputSize));
%    
%   % Apply OFDM modulation to the random inputs
%   modOut = step(hMod, dataIn, pilotIn);
%   % Pass the modulator output through a 2x4 static, noiseless channel
%   chanOut = modOut * ones(2, 4);
%   % Apply OFDM demodulation
%   [dataOut, pilotOut] = step(hDemod, chanOut);
%    
%   % Retrieve the pilot signals from the demodulator output at any receive
%   % antenna because null subcarriers are used to avoid interference
%   % between the pilots from different transmit antennas 
%   pilotOutAtOneRx = pilotOut(:,:,:,1); % = pilotOut(:,:,:,i), i = 2,3,4
%   isSame = (max(abs(pilotOutAtOneRx(:) - pilotIn(:))) < 1e-10)
% 
%   See also comm.OFDMModulator.

%   Copyright 2013 The MathWorks, Inc.

%#codegen
%#ok<*EMCA>

properties (Nontunable)
    %NumReceiveAntennas Number of receive antennas
    %   Specify the number of receive antennas as a numeric, real,
    %   positive integer scalar between 1 and 64, inclusive. The default
    %   value of this property is 1.
    NumReceiveAntennas = 1;
end

properties (Nontunable, Logical)
    %RemoveDCCarrier Remove DC subcarrier from output
    %   Set this property to true to remove the direct current (DC)
    %   subcarrier from the output for each OFDM symbol. The DC subcarrier
    %   is the center of the frequency band and indexed at FFTLength/2 + 1.
    %   The default value of this property is false.
    RemoveDCCarrier = false;
    %PilotOutputPort Pilot output port
    %   Set this property to true to enable a secondary output for pilot
    %   signals, which are from the subcarriers specified by the
    %   PilotCarrierIndices property. The default value of this property is
    %   false.
    PilotOutputPort = false;
end

methods
  function object = DOFDMDemodulator(varargin)
    if nargin == 1 && isa(varargin{1}, 'comm.OFDMModulator')
        modObj = varargin{1};        
        object.FFTLength            = modObj.FFTLength;
        object.NumGuardBandCarriers = modObj.NumGuardBandCarriers;     
        object.CyclicPrefixLength   = modObj.CyclicPrefixLength;
        object.RemoveDCCarrier      = modObj.InsertDCNull;
        object.NumSymbols           = modObj.NumSymbols;
        object.PilotOutputPort      = modObj.PilotInputPort;
        if modObj.PilotInputPort
            object.PilotCarrierIndices = modObj.PilotCarrierIndices;
        end        
    else
        setProperties(object, nargin, varargin{:});
    end
  end

  function set.NumReceiveAntennas(obj, Nr)
    propName = 'NumReceiveAntennas';
    validateattributes(Nr, {'numeric'}, ... 
        {'real','scalar','integer','nonempty','>=',1,'<=',64}, ... 
        [class(obj) '.' propName], propName); 
    
    obj.NumReceiveAntennas = Nr;
  end
  
  function showResourceMapping(obj, varargin)
    %showResourceMapping Display subcarrier mapping for data, pilot, guard band and null signals
    %   showResourceMapping(H) displays the subcarrier mapping for data,
    %   pilot, guard band and null signals for each symbol and each
    %   transmit antenna, based on the configuration of the OFDM
    %   demodulator System object H. Different signal types are shown in
    %   different colors as indicated in the colorbar on each plot.
    %   
    %   If the subcarrier mapping is different across transmit antennas,
    %   i.e., when the PilotCarrierIndices property of H applies and is a
    %   3D array, each transmit antenna has its own mapping plot for the
    %   Nsym symbols, where Nsym is the NumSymbols property value of H.
    %   Otherwise, only one mapping plot is displayed for all the transmit
    %   antennas. In each plot, the x-axis indicates the Nsym OFDM symbols
    %   and the y-axis indicates the NFFT subcarriers that are indexed from
    %   1 to NFFT, where NFFT is the FFTLength property value of H.
    %   
    %   This method can display up to 16 subcarrier mapping plots per call.
    % 
    %   showResourceMapping(H, freqIndex) customizes the y-axis frequency
    %   subcarrier index by freqIndex. freqIndex must be numeric, real,
    %   integer row vector of length 2 such that freqIndex(2) -
    %   freqIndex(1) = FFTLen - 1. The frequency subcarriers are indexed
    %   from freqIndex(1) to freqIndex(2).
    %
    %   showResourceMapping(H, freqIndex, txIndex) displays subcarrier
    %   mapping plots for selected transmit antennas indicated by txIndex.
    %   This syntax applies only when each transmit antenna has its own
    %   mapping plot, i.e., when the PilotCarrierIndices property of H is a
    %   3D array. txIndex is a numeric, real, integer scalar or row vector
    %   with values between 1 and Nt, inclusive, where Nt is the number of
    %   transmit antennas, i.e., the third dimension size of the
    %   PilotCarrierIndices property value of H. The number of unique
    %   elements in txIndex must be less than or equal to 16.
    
    coder.internal.errorIf(~isempty(coder.target), 'comm:OFDM:NoCodegenForResourceMapping');
    hMod = comm.OFDMModulator(obj);
    hMod.showResourceMapping(varargin{:});
  end
end

methods(Access = protected)
  function validatePropertiesImpl(obj)          
    dependentPropertyValidation(obj, obj.PilotOutputPort, obj.RemoveDCCarrier);
  end
  
  function validateInputsImpl(obj, x)
    validateattributes(x, {'double'}, ...
        {'2d','nonempty','finite','size', [getTimeDomainFrameLength(obj) obj.NumReceiveAntennas]}, ...
        [class(obj) '.' 'Input'], 'Input');
  end
  
  function setupImpl(obj, varargin)        
    obj.pHasPilot      = obj.PilotOutputPort;
    obj.pHasDCNull     = obj.RemoveDCCarrier;
    obj.pNumAntennas   = obj.NumReceiveAntennas;    
    obj.pHasCustomNull = obj.PilotOutputPort && ~ismatrix(obj.PilotCarrierIndices);
    
    if ((floor(log2(obj.FFTLength)) ~= log2(obj.FFTLength))) && ~isempty(coder.target)
        % Use dsp.FFT System object for non-power-of-2 FFT in code generation
        obj.pFFTFcnPtr = @FFTWithSysObj;
    else
        obj.pFFTFcnPtr = @FFTWithFcn;
    end
    
    if obj.PilotOutputPort % Calculate pilot linear indices
        obj.pPilotLinearIndices = ...
            getPilotLinearIndex(obj, obj.pHasCustomNull);
    end
    
    % Calculate data linear indices
    getDataLinearIndex(obj, false);
  end
  
  function varargout = stepImpl(obj, x)
    FFTLen = obj.FFTLength;
    CPLen  = obj.CyclicPrefixLength;
    numSym = obj.NumSymbols;
    numRx  = obj.NumReceiveAntennas;        
      
    % Remove cyclic prefix
    if isscalar(CPLen)
        inputIn3D = reshape(x, [(FFTLen + CPLen) numSym numRx]);
        postCPRemoval = inputIn3D(CPLen+1:end, :, :);
    else
        postCPRemoval = coder.nullcopy(complex(zeros(FFTLen, numSym, numRx)));
        for symIdx = 1:numSym
            postCPRemoval(:, symIdx, :) = x(FFTLen*(symIdx-1)+sum(CPLen(1:symIdx))+(1:FFTLen), :);
        end
    end

    % FFT
    postFFT = obj.pFFTFcnPtr(postCPRemoval);    

    
    % FFT shift
    if isreal(postFFT)
        postShift = complex(fftshift(postFFT, 1), 0);
    else
        postShift = fftshift(postFFT,1);
    end
    
    % Output data
    varargout{1} = reshape(postShift(obj.pDataLinearIndices), ...
        [length(obj.pDataLinearIndices)/numSym/numRx numSym numRx]);
    
    % Output pilot
    if obj.PilotOutputPort
        if obj.pHasCustomNull % In the format of [numPilot numSym numTx numRx]
            numTx = size(obj.PilotCarrierIndices, 3);
            varargout{2} = permute(reshape(postShift(obj.pPilotLinearIndices), ...
                [length(obj.pPilotLinearIndices)/numTx/numSym/numRx numTx numSym numRx]), ...
                [1 3 2 4]);
        else % In the format of [numPilot numSym numRx]
            varargout{2} = reshape(postShift(obj.pPilotLinearIndices), ...
                [length(obj.pPilotLinearIndices)/numSym/numRx numSym numRx]);
        end
    end
  end
  
  function flag = isInactivePropertyImpl(obj, prop)
    switch prop                    
    case 'PilotCarrierIndices'
        flag = ~(obj.PilotOutputPort);        
    otherwise
        flag = false;
    end
  end
  
  function s = saveObjectImpl(obj)
    s = saveObjectImpl@comm.internal.OFDMBase(obj);
  end

  function loadObjectImpl(obj, s, wasLocked)
    loadObjectImpl@comm.internal.OFDMBase(obj, s, wasLocked);
  end

  function s = infoImpl(obj)          
    %info Returns input and output size information of the OFDM demodulator object
    %   S = info(H) returns a structure, S, containing the input and output
    %   size information of the OFDM demodulator System object, H. The
    %   structure S has the following fields:
    % 
    %   InputSize       - Signal input size
    %   DataOutputSize  - Data signal output size
    %   PilotOutputSize - Pilot signal output size (This field exists when 
    %                     the PilotOutputPort property is set to true)
    
    if ~isLocked(obj)        
        validateProperties(obj);
    end
    
    numSym = obj.NumSymbols;
    numRx  = obj.NumReceiveAntennas;
    
    s.InputSize = [getTimeDomainFrameLength(obj) numRx];

    if numRx == 1
        s.DataOutputSize = [getNumDataCarriers(obj, obj.RemoveDCCarrier, obj.PilotOutputPort) numSym];
    else
        s.DataOutputSize = [getNumDataCarriers(obj, obj.RemoveDCCarrier, obj.PilotOutputPort) numSym numRx];
    end
    
    if obj.PilotOutputPort
        if ismatrix(obj.PilotCarrierIndices)
            if numRx == 1
                s.PilotOutputSize = [size(obj.PilotCarrierIndices, 1) numSym];
            else
                s.PilotOutputSize = [size(obj.PilotCarrierIndices, 1) numSym numRx];
            end
        else
            if numRx == 1
                s.PilotOutputSize = [size(obj.PilotCarrierIndices, 1) numSym size(obj.PilotCarrierIndices, 3)];
            else
                s.PilotOutputSize = [size(obj.PilotCarrierIndices, 1) numSym size(obj.PilotCarrierIndices, 3) numRx];
            end
        end        
    end
  end
  
  function num = getNumOutputsImpl(obj)
    num = 1 + obj.PilotOutputPort;
  end

  function flag = isInputSizeLockedImpl(~,~)
     flag = true;
  end
  
  function flag = isInputComplexityLockedImpl(~,~)
    flag = false;
  end
  
  function flag = isOutputComplexityLockedImpl(~,~)
    flag = true;
  end
    
  function varargout = isOutputFixedSizeImpl(obj)
    if obj.PilotOutputPort
        varargout = {true, true};
    else
        varargout = {true};
    end
  end
  
  function varargout = isOutputComplexImpl(obj)
    if obj.PilotOutputPort
        varargout = {true, true};
    else
        varargout = {true};
    end
  end

  function varargout = getOutputDataTypeImpl(obj)
    if obj.PilotOutputPort
        varargout = {'double', 'double'};
    else
        varargout = {'double'};
    end
  end

  function varargout = getOutputSizeImpl(obj)
    numSym = obj.NumSymbols;
    numRx  = obj.NumReceiveAntennas;

    varargout = {[getNumDataCarriers(obj, obj.RemoveDCCarrier, obj.PilotOutputPort) numSym numRx]};

    if (obj.PilotOutputPort)
        if ismatrix(obj.PilotCarrierIndices)
            varargout{end+1} = [size(obj.PilotCarrierIndices, 1) numSym numRx];
        else
            varargout{end+1} = [size(obj.PilotCarrierIndices, 1) numSym ...
                                size(obj.PilotCarrierIndices, 3) numRx];
        end
    end
  end
    
  function varargout = getInputNamesImpl(~)
    varargout = {''};
  end
  
  function varargout = getOutputNamesImpl(obj)
    if obj.PilotOutputPort
        varargout = {'', 'Pilot'}; 
    else
        varargout = {''}; 
    end    
  end
  
  function icon = getIconImpl(~)
    icon = sprintf('OFDM\nDemodulator');
  end  
end

methods(Static, Access=protected)
  function header = getHeaderImpl    
    header = matlab.system.display.Header('comm.OFDMDemodulator',...
        'Title', 'OFDM Demodulator', 'Text', ...
        ['Apply OFDM demodulation to the input signal. Enable pilot signal ' sprintf('\n'), ...
         'output to separate it from the data signal after demodulation.']);
  end
  
  function groups = getPropertyGroupsImpl
    params = matlab.system.display.Section(...
        'Title', 'Parameters',...
        'PropertyList', {'FFTLength','NumGuardBandCarriers','RemoveDCCarrier',...
                         'PilotOutputPort','PilotCarrierIndices',...
                         'CyclicPrefixLength','NumSymbols','NumReceiveAntennas'});
    groups = params;
  end
end

end

function y = FFTWithFcn(x)
    y = fft(x, [], 1);
end
function y = FFTWithSysObj(x) % Only for code generation
    hFFT = dsp.FFT;
    y = step(hFFT, x);
end

% [EOF]