classdef DOFDMTransmitterBits < matlab.System
% OFDMTransmitter Generate OFDM signal repeatedly for a payload message
% based on the 802.11a standard.

% Copyright 2014 The MathWorks, Inc.

properties (Nontunable)
    %SampleRate Sample rate    
    SampleRate = 20e6
    %PayloadMessage Payload message
    PayloadBits = repmat([0 1 0 1 0 1 1 1 1 1 0 0 0 1 0 1 0 1 0 0 0 1 0 0 1 1 1 1 0 1 1 0 1 0 0 1 1 1 0 1 0 0 0 0 1 0 1 0],1, 500);
    %NumFrames Number of frames to transmit
    NumFrames = 2
end

properties(Nontunable, SetAccess = private, Dependent)
    %FrameLength Frame length
    FrameLength
end

properties(Nontunable, Access = private)
    pPayloadBits        % Number of bits from payload message
    pNumPadBits         % Number of random bits padded in each frame
    pNumOFDMSymbols     % Number of OFDM symbols per frame
    pPreamble           % Preamble for each frame
    pPilots             % Pilots for each frame
    pBPSKMod            % BPSK Modulator System object
    pDataOFDMMod        % OFDM Modulator System object 
end

properties(Constant, Access = private) % OFDM Modulator configuration
    FFTLength            = 64
    NumGuardBandCarriers = [6; 5]
    NumDataCarriers      = 48
    CyclicPrefixLength   = 16
    PilotCarrierIndices  = [12;26;40;54]
    NumOFDMSymInPreamble = 5
end

methods        
  function obj = DOFDMTransmitterBits(varargin)
    setProperties(obj, nargin, varargin{:});
  end
  
  function set.SampleRate(obj, Rs)
    propName = 'SampleRate';
    validateattributes(Rs, {'double'}, ...
        {'real','scalar','positive','finite'}, ...
        [class(obj) '.' propName], propName); 
    
	obj.SampleRate = Rs;
  end
  
  function set.PayloadBits(obj, msg)
    propName = 'PayloadBits';
    validateattributes(msg, {'double'}, {'row'}, ... 
        [class(obj) '.' propName], propName); 
    
    obj.PayloadMessage = msg;
  end
  
  function set.NumFrames(obj, numFrm)
    propName = 'NumFrames';
    validateattributes(numFrm, {'double'}, ... 
        {'real','scalar','positive','integer','finite'}, ... 
        [class(obj) '.' propName], propName); 
    
    obj.NumFrames = numFrm;
  end
  
  function frameLen = get.FrameLength(obj)
    frameLen = ceil(length(obj.PayloadBits)/obj.NumDataCarriers) * ...
               (obj.FFTLength+obj.CyclicPrefixLength) + obj.NumOFDMSymInPreamble*obj.FFTLength;
  end
end

methods(Access = protected)
  function setupImpl(obj)
    coder.extrinsic('dec2bin','getOFDMPreambleAndPilot');
    
    % Convert message to bits
    % msgInBits = coder.const(double(dec2bin(obj.PayloadMessage, obj.NumBitsPerCharacter).'));
    obj.pPayloadBits = obj.PayloadBits';

    % Calculate number of OFDM symbols per frame
    obj.pNumOFDMSymbols = ceil(length(obj.pPayloadBits)/obj.NumDataCarriers);
    % Calculate number of bits padded in each frame
    obj.pNumPadBits = obj.NumDataCarriers * obj.pNumOFDMSymbols - length(obj.pPayloadBits);
    
    % Get preamble for each frame
    obj.pPreamble = coder.const(double(getOFDMPreambleAndPilot('Preamble', obj.FFTLength, obj.NumGuardBandCarriers)));

    % Get pilot for each frame
    obj.pPilots = coder.const(double(getOFDMPreambleAndPilot('Pilot', obj.pNumOFDMSymbols)));

    % BPSK modulator
    obj.pBPSKMod = comm.BPSKModulator;

    % OFDM modulator
    obj.pDataOFDMMod = comm.OFDMModulator(...
        'FFTLength' ,           obj.FFTLength, ...
        'NumGuardBandCarriers', obj.NumGuardBandCarriers, ...
        'InsertDCNull',         true, ...
        'PilotInputPort',       true, ...
        'PilotCarrierIndices',  obj.PilotCarrierIndices, ...
        'CyclicPrefixLength',   obj.CyclicPrefixLength, ...
        'NumSymbols',           obj.pNumOFDMSymbols);
    
    global mod
    mod = obj.pDataOFDMMod;
    
  end
  
  function resetImpl(obj)
    reset(obj.pBPSKMod);
    reset(obj.pDataOFDMMod);
  end
  
  function y = stepImpl(obj)    
    % BPSK modulation for one frame
    symPostBPSK = step(obj.pBPSKMod, ...
        [obj.pPayloadBits; randi([0 1], obj.pNumPadBits, 1)]);
  
    % OFDM modulation for one frame
    symPostOFDM = step(obj.pDataOFDMMod, ...
        reshape(symPostBPSK, obj.NumDataCarriers, obj.pNumOFDMSymbols), obj.pPilots);
    
    % Repeat the frame
    y = repmat([obj.pPreamble; symPostOFDM], obj.NumFrames, 1);
  end
  
  function releaseImpl(obj)
    release(obj.pBPSKMod);
    release(obj.pDataOFDMMod);
  end
  
  function s = saveObjectImpl(obj)
    s = saveObjectImpl@matlab.System(obj);
    if isLocked(obj)
        s.pBPSKMod        = matlab.System.saveObject(obj.pBPSKMod);
        s.pDataOFDMMod    = matlab.System.saveObject(obj.pDataOFDMMod);
        s.pPayloadBits    = obj.pPayloadBits;
        s.pNumPadBits     = obj.pNumPadBits;
        s.pNumOFDMSymbols = obj.pNumOFDMSymbols;
        s.pPreamble       = obj.pPreamble;
        s.pPilots         = obj.pPilots;
    end      
  end
  
  function loadObjectImpl(obj, s, wasLocked)
    if wasLocked
        obj.pBPSKMod        = matlab.System.loadObject(s.pBPSKMod);
        obj.pDataOFDMMod    = matlab.System.loadObject(s.pDataOFDMMod);
        obj.pPayloadBits    = s.pPayloadBits;
        obj.pNumPadBits     = s.pNumPadBits;
        obj.pNumOFDMSymbols = s.pNumOFDMSymbols;
        obj.pPreamble       = s.pPreamble;
        obj.pPilots         = s.pPilots;
    end
    loadObjectImpl@matlab.System(obj, s);
  end

  function num = getNumInputsImpl(~)
    num = 0;
  end
    
  function flag = isOutputComplexityLockedImpl(~,~)
    flag = true;
  end
end

end

% [EOF]